@@ -26,6 +26,8 @@
 #include "lttng-events.h"
 #include "lttng-tracer.h"
 
+static struct lttng_transport lttng_relay_transport;
+
 struct metadata_packet_header {
 	uint32_t magic;			/* 0x75D11D57 */
 	uint8_t  uuid[16];		/* Unique Universal Identifier */
@@ -148,6 +150,54 @@ static void client_buffer_finalize(struct lib_ring_buffer *buf, void *priv, int
 {
 }
 
+static int client_timestamp_begin(const struct lib_ring_buffer_config *config,
+		struct lib_ring_buffer *buf, uint64_t *timestamp_begin)
+{
+	return -ENOSYS;
+}
+
+static int client_timestamp_end(const struct lib_ring_buffer_config *config,
+			struct lib_ring_buffer *bufb,
+			uint64_t *timestamp_end)
+{
+	return -ENOSYS;
+}
+
+static int client_events_discarded(const struct lib_ring_buffer_config *config,
+			struct lib_ring_buffer *bufb,
+			uint64_t *events_discarded)
+{
+	return -ENOSYS;
+}
+
+static int client_current_timestamp(const struct lib_ring_buffer_config *config,
+		struct lib_ring_buffer *bufb,
+		uint64_t *ts)
+{
+	return -ENOSYS;
+}
+
+static int client_content_size(const struct lib_ring_buffer_config *config,
+			struct lib_ring_buffer *bufb,
+			uint64_t *content_size)
+{
+	return -ENOSYS;
+}
+
+static int client_packet_size(const struct lib_ring_buffer_config *config,
+			struct lib_ring_buffer *bufb,
+			uint64_t *packet_size)
+{
+	return -ENOSYS;
+}
+
+static int client_stream_id(const struct lib_ring_buffer_config *config,
+			struct lib_ring_buffer *bufb,
+			uint64_t *stream_id)
+{
+	return -ENOSYS;
+}
+
 static const struct lib_ring_buffer_config client_config = {
 	.cb.ring_buffer_clock_read = client_ring_buffer_clock_read,
 	.cb.record_header_size = client_record_header_size,
@@ -169,15 +219,9 @@ static const struct lib_ring_buffer_config client_config = {
 };
 
 static
-struct channel *_channel_create(const char *name,
-				struct lttng_channel *lttng_chan, void *buf_addr,
-				size_t subbuf_size, size_t num_subbuf,
-				unsigned int switch_timer_interval,
-				unsigned int read_timer_interval)
+void release_priv_ops(void *priv_ops)
 {
-	return channel_create(&client_config, name, lttng_chan, buf_addr,
-			      subbuf_size, num_subbuf, switch_timer_interval,
-			      read_timer_interval);
+	module_put(THIS_MODULE);
 }
 
 static
@@ -186,6 +230,37 @@ void lttng_channel_destroy(struct channel *chan)
 	channel_destroy(chan);
 }
 
+static
+struct channel *_channel_create(const char *name,
+				struct lttng_channel *lttng_chan, void *buf_addr,
+				size_t subbuf_size, size_t num_subbuf,
+				unsigned int switch_timer_interval,
+				unsigned int read_timer_interval)
+{
+	struct channel *chan;
+
+	chan = channel_create(&client_config, name, lttng_chan, buf_addr,
+			      subbuf_size, num_subbuf, switch_timer_interval,
+			      read_timer_interval);
+	if (chan) {
+		/*
+		 * Ensure this module is not unloaded before we finish
+		 * using lttng_relay_transport.ops.
+		 */
+		if (!try_module_get(THIS_MODULE)) {
+			printk(KERN_WARNING "LTT : Can't lock transport module.\n");
+			goto error;
+		}
+		chan->backend.priv_ops = &lttng_relay_transport.ops;
+		chan->backend.release_priv_ops = release_priv_ops;
+	}
+	return chan;
+
+error:
+	lttng_channel_destroy(chan);
+	return NULL;
+}
+
 static
 struct lib_ring_buffer *lttng_buffer_read_open(struct channel *chan)
 {
@@ -240,7 +315,7 @@ static
 void lttng_event_write_from_user(struct lib_ring_buffer_ctx *ctx,
 			       const void __user *src, size_t len)
 {
-	lib_ring_buffer_copy_from_user(&client_config, ctx, src, len);
+	lib_ring_buffer_copy_from_user_inatomic(&client_config, ctx, src, len);
 }
 
 static
@@ -250,6 +325,13 @@ void lttng_event_memset(struct lib_ring_buffer_ctx *ctx,
 	lib_ring_buffer_memset(&client_config, ctx, c, len);
 }
 
+static
+void lttng_event_strcpy(struct lib_ring_buffer_ctx *ctx, const char *src,
+		size_t len)
+{
+	lib_ring_buffer_strcpy(&client_config, ctx, src, len, '#');
+}
+
 static
 size_t lttng_packet_avail_size(struct channel *chan)
 			     
@@ -308,11 +390,19 @@ static struct lttng_transport lttng_relay_transport = {
 		.event_write_from_user = lttng_event_write_from_user,
 		.event_memset = lttng_event_memset,
 		.event_write = lttng_event_write,
+		.event_strcpy = lttng_event_strcpy,
 		.packet_avail_size = lttng_packet_avail_size,
 		.get_writer_buf_wait_queue = lttng_get_writer_buf_wait_queue,
 		.get_hp_wait_queue = lttng_get_hp_wait_queue,
 		.is_finalized = lttng_is_finalized,
 		.is_disabled = lttng_is_disabled,
+		.timestamp_begin = client_timestamp_begin,
+		.timestamp_end = client_timestamp_end,
+		.events_discarded = client_events_discarded,
+		.content_size = client_content_size,
+		.packet_size = client_packet_size,
+		.stream_id = client_stream_id,
+		.current_timestamp = client_current_timestamp,
 	},
 };
 
