@@ -35,16 +35,18 @@
 #include <asm/local.h>
 
 #include "wrapper/trace-clock.h"
+#include "wrapper/compiler.h"
 #include "lttng-tracer-core.h"
 #include "lttng-events.h"
 
 #define LTTNG_MODULES_MAJOR_VERSION 2
-#define LTTNG_MODULES_MINOR_VERSION 0
-#define LTTNG_MODULES_PATCHLEVEL_VERSION 2
+#define LTTNG_MODULES_MINOR_VERSION 5
+#define LTTNG_MODULES_PATCHLEVEL_VERSION 0
+#define LTTNG_MODULES_EXTRAVERSION ""
 
-#define LTTNG_VERSION_NAME		"Annedd'ale"
+#define LTTNG_VERSION_NAME		"Fumisterie"
 #define LTTNG_VERSION_DESCRIPTION	\
-	"New type of beer, 100% from Quebec, flavored with sapin beaumier needles, with a touch of hops."
+	"The Fumisterie is a beer whose style is at the confluence of English and German ales. On the palate, it is full-bodied, malty and slightly caramelized. The organic hemp seeds incorporated during the brewing gives it a unique and original taste. The hop flavour comes through delicately in the after taste."
 
 #ifndef CHAR_BIT
 #define CHAR_BIT 8
